#pragma once

#include "myirc/connection.hpp"
#include "myirc/sasl_mechanism.hpp"

#include <string>
#include <unordered_set>
#include <span>

namespace myirc {

struct IrcMsg;

enum class Casemap
{
    Rfc1459,
    Rfc1459_Strict,
    Ascii,
};

struct Chat {
    std::span<const irctag> tags;
    bool is_notice;
    char status_msg;
    std::string_view source;
    std::string_view target;
    std::string_view message;
};

/**
 * @brief Thread to track this connection's identity, and IRC state.
 *
 */
class Client
{
    std::shared_ptr<Connection> connection_;

    std::string nickname_;
    std::string mode_;
    std::unordered_set<std::string> channels_;

    // RPL_ISUPPORT state
    std::unordered_map<std::string, std::string> isupport_;
    std::unique_ptr<SaslMechanism> sasl_mechanism_;

    Casemap casemap_;
    std::string channel_prefix_;
    std::string status_msg_;

    std::unordered_map<std::string, std::string> caps_available_;
    std::unordered_set<std::string> caps_;

    // AUTHENTICATE support
    std::string authenticate_buffer_;

    auto on_authenticate(std::string_view) -> void;
    auto on_authenticate_chunk(std::string_view) -> void;

    auto on_welcome(const IrcMsg &irc) -> void;
    auto on_isupport(const IrcMsg &irc) -> void;
    auto on_nick(const IrcMsg &irc) -> void;
    auto on_umodeis(const IrcMsg &irc) -> void;
    auto on_join(const IrcMsg &irc) -> void;
    auto on_kick(const IrcMsg &irc) -> void;
    auto on_part(const IrcMsg &irc) -> void;
    auto on_mode(const IrcMsg &irc) -> void;
    auto on_cap(const IrcMsg &irc) -> void;
    auto on_registered() -> void;
    auto on_chat(bool, const IrcMsg &irc, bool flush) -> void;

public:
    boost::signals2::signal<void()> sig_registered;
    boost::signals2::signal<void(const std::unordered_map<std::string, std::string> &)> sig_cap_ls;
    boost::signals2::signal<void(const Chat &, bool flush)> sig_chat;
    boost::signals2::signal<void(SnoteMatch &, bool flush)> sig_snote;

    Client(std::shared_ptr<Connection> connection)
        : connection_{std::move(connection)}
        , casemap_{Casemap::Rfc1459}
        , channel_prefix_{"#&"}
        , status_msg_{"+@"}
    {
    }

    auto get_connection() -> Connection & { return *connection_; }

    static auto start(std::shared_ptr<Connection>) -> std::shared_ptr<Client>;

    auto start_sasl(std::unique_ptr<SaslMechanism> mechanism) -> void;

    auto get_connection() const -> std::shared_ptr<Connection>
    {
        return connection_->shared_from_this();
    }

    auto get_my_nick() const -> const std::string &;
    auto get_my_mode() const -> const std::string &;
    auto get_my_channels() const -> const std::unordered_set<std::string> &;

    auto list_caps() -> void;

    auto is_my_nick(std::string_view nick) const -> bool;
    auto is_my_mask(std::string_view mask) const -> bool;
    auto is_channel(std::string_view name) const -> bool;

    auto casemap(std::string_view) const -> std::string;
    auto casemap_compare(std::string_view, std::string_view) const -> int;

    auto shutdown() -> void;

    auto send_ping(std::string_view) -> void;
    auto send_pong(std::string_view) -> void;
    auto send_pass(std::string_view) -> void;
    auto send_user(std::string_view, std::string_view) -> void;
    auto send_nick(std::string_view) -> void;
    auto send_join(std::string_view) -> void;
    auto send_names(std::string_view channel) -> void;
    auto send_kick(std::string_view, std::string_view, std::string_view) -> void;
    auto send_kill(std::string_view, std::string_view) -> void;
    auto send_quit(std::string_view) -> void;
    auto send_cap_ls() -> void;
    auto send_cap_end() -> void;
    auto send_map() -> void;
    auto send_testline(std::string_view) -> void;
    auto send_testmask(std::string_view) -> void;
    auto send_testmask_gecos(std::string_view, std::string_view) -> void;
    auto send_masktrace(std::string_view) -> void;
    auto send_masktrace_gecos(std::string_view, std::string_view) -> void;
    auto send_get_topic(std::string_view) -> void;
    auto send_set_topic(std::string_view, std::string_view) -> void;
    auto send_cap_req(std::string_view) -> void;
    auto send_privmsg(std::string_view, std::string_view) -> void;
    auto send_wallops(std::string_view) -> void;
    auto send_notice(std::string_view, std::string_view) -> void;
    auto send_authenticate(std::string_view message) -> void;
    auto send_authenticate_encoded(std::string_view message) -> void;
    auto send_authenticate_abort() -> void;
    auto send_whois(std::string_view) -> void;
    auto send_whois_remote(std::string_view, std::string_view) -> void;
    auto send_challenge(std::string_view) -> void;
    auto send_oper(std::string_view, std::string_view) -> void;
};

} // namespace myirc
